/*
 * Written by Alexey Vatchenko <av@bsdua.org>.
 * Public domain.
 */
#include <sys/types.h>
#include <sys/ioctl.h>

#include <machine/apmvar.h>

#include <gtk/gtk.h>

#include <err.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include "tray-app.h"
#include "battery-connected.xpm"
#include "battery-full.xpm"
#include "battery-empty.xpm"
#include "battery-charge.xpm"
#include "battery-90.xpm"
#include "battery-80.xpm"
#include "battery-70.xpm"
#include "battery-60.xpm"
#include "battery-50.xpm"
#include "battery-40.xpm"
#include "battery-30.xpm"
#include "battery-20.xpm"
#include "battery-10.xpm"

static void	usage(const char *prog);
static int	get_power(int fd, struct apm_power_info *batinfo);
static int	is_equal_info(struct apm_power_info *info1,
		    struct apm_power_info *info2);

static void	prepare_tooltip(struct apm_power_info *info, char *text,
		    size_t sz);

static gboolean	cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
		    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data);

static int	init_tray(int doinvert);
static void	set_tray_icon(struct apm_power_info *info);

static int			apm_fd;
static struct apm_power_info	current_info;

static GtkStatusIcon	*tray_icon = NULL;
static GdkPixbuf	*tray_connected_pixbuf = NULL;
static GdkPixbuf	*tray_full_pixbuf = NULL;
static GdkPixbuf	*tray_empty_pixbuf = NULL;
static GdkPixbuf	*tray_charge_pixbuf = NULL;
static GdkPixbuf	*tray_90_pixbuf = NULL;
static GdkPixbuf	*tray_80_pixbuf = NULL;
static GdkPixbuf	*tray_70_pixbuf = NULL;
static GdkPixbuf	*tray_60_pixbuf = NULL;
static GdkPixbuf	*tray_50_pixbuf = NULL;
static GdkPixbuf	*tray_40_pixbuf = NULL;
static GdkPixbuf	*tray_30_pixbuf = NULL;
static GdkPixbuf	*tray_20_pixbuf = NULL;
static GdkPixbuf	*tray_10_pixbuf = NULL;

void
usage(const char *prog)
{
	printf("Usage:\t%s [-h] [-i]\n"
	    "Options:\n"
	    "\t -h              - print this help\n"
	    "\t -i              - invert foreground color\n",
	    prog);
}

static int
get_power(int fd, struct apm_power_info *info)
{
	memset(info, 0, sizeof(*info));
	return (ioctl(fd, APM_IOC_GETPOWER, info));
}

static int
is_equal_info(struct apm_power_info *info1, struct apm_power_info *info2)
{
	return (info1->ac_state == info2->ac_state &&
	    info1->battery_life == info2->battery_life);
}

static void
prepare_tooltip(struct apm_power_info *info, char *text, size_t sz)
{
	char *s;

	if (info->ac_state == APM_AC_ON) {
		if (info->battery_life == 100)
			s = "External power";
		else
			s = "Charging";
	} else
		s = "Battery";

	snprintf(text, sz, "%s: %u%%", s, (u_int)info->battery_life);
}

static gboolean
cb_timer(GtkWidget *widget)
{
	struct apm_power_info info;

	if (get_power(apm_fd, &info) == 0 &&
	    !is_equal_info(&info, &current_info)) {
		set_tray_icon(&info);
		memcpy(&current_info, &info, sizeof(current_info));
	}
	return (TRUE);
}

static gboolean
cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data)
{
	char text[30];

	prepare_tooltip(&current_info, text, sizeof(text));
	gtk_tooltip_set_text(tooltip, text);
	return (TRUE);
}

static int
init_tray(int doinvert)
{
	struct apm_power_info info;
	char tooltip[30];

	tray_connected_pixbuf = gdk_pixbuf_new_from_xpm_data(
	    battery_connected_xpm);
	if (tray_connected_pixbuf == NULL)
		return (-1);
	tray_full_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_full_xpm);
	if (tray_full_pixbuf == NULL)
		return (-1);
	tray_empty_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_empty_xpm);
	if (tray_empty_pixbuf == NULL)
		return (-1);
	tray_charge_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_charge_xpm);
	if (tray_charge_pixbuf == NULL)
		return (-1);
	tray_90_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_90_xpm);
	if (tray_90_pixbuf == NULL)
		return (-1);
	tray_80_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_80_xpm);
	if (tray_80_pixbuf == NULL)
		return (-1);
	tray_70_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_70_xpm);
	if (tray_70_pixbuf == NULL)
		return (-1);
	tray_60_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_60_xpm);
	if (tray_60_pixbuf == NULL)
		return (-1);
	tray_50_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_50_xpm);
	if (tray_50_pixbuf == NULL)
		return (-1);
	tray_40_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_40_xpm);
	if (tray_40_pixbuf == NULL)
		return (-1);
	tray_30_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_30_xpm);
	if (tray_30_pixbuf == NULL)
		return (-1);
	tray_20_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_20_xpm);
	if (tray_20_pixbuf == NULL)
		return (-1);
	tray_10_pixbuf = gdk_pixbuf_new_from_xpm_data(battery_10_xpm);
	if (tray_10_pixbuf == NULL)
		return (-1);

	if (doinvert) {
		trayapp_invert_pixbuf(tray_connected_pixbuf);
		trayapp_invert_pixbuf(tray_full_pixbuf);
		trayapp_invert_pixbuf(tray_empty_pixbuf);
		trayapp_invert_pixbuf(tray_charge_pixbuf);
		trayapp_invert_pixbuf(tray_90_pixbuf);
		trayapp_invert_pixbuf(tray_80_pixbuf);
		trayapp_invert_pixbuf(tray_70_pixbuf);
		trayapp_invert_pixbuf(tray_60_pixbuf);
		trayapp_invert_pixbuf(tray_50_pixbuf);
		trayapp_invert_pixbuf(tray_40_pixbuf);
		trayapp_invert_pixbuf(tray_30_pixbuf);
		trayapp_invert_pixbuf(tray_20_pixbuf);
		trayapp_invert_pixbuf(tray_10_pixbuf);
	}

	tray_icon = gtk_status_icon_new();
	if (tray_icon == NULL)
		return (-1);
	get_power(apm_fd, &info);
	set_tray_icon(&info);
	memcpy(&current_info, &info, sizeof(current_info));

	prepare_tooltip(&info, tooltip, sizeof(tooltip));
	gtk_status_icon_set_tooltip_text(tray_icon, tooltip);

	g_signal_connect(tray_icon, "query-tooltip",
	    G_CALLBACK(cb_tray_query_tooltip), NULL);
	return (0);
}

static void
set_tray_icon(struct apm_power_info *info)
{
	GdkPixbuf *pb;

	if (info->ac_state == APM_AC_ON) {
		if (info->battery_life == 100)
			pb = tray_connected_pixbuf;
		else
			pb = tray_charge_pixbuf;
	} else if (info->battery_life >= 95)
		pb = tray_full_pixbuf;
	else if (info->battery_life >= 85)
		pb = tray_90_pixbuf;
	else if (info->battery_life >= 75)
		pb = tray_80_pixbuf;
	else if (info->battery_life >= 65)
		pb = tray_70_pixbuf;
	else if (info->battery_life >= 55)
		pb = tray_60_pixbuf;
	else if (info->battery_life >= 45)
		pb = tray_50_pixbuf;
	else if (info->battery_life >= 35)
		pb = tray_40_pixbuf;
	else if (info->battery_life >= 25)
		pb = tray_30_pixbuf;
	else if (info->battery_life >= 15)
		pb = tray_20_pixbuf;
	else if (info->battery_life >= 5)
		pb = tray_10_pixbuf;
	else
		pb = tray_empty_pixbuf;

	gtk_status_icon_set_from_pixbuf(tray_icon, pb);
}

int
main(int argc, char **argv)
{
	char *progname;
	int ch, invert_flag;
	int error;

	progname = argv[0];

	if (!gtk_init_check(&argc, &argv))
		errx(1, "Cannot initialize windowing system");

	invert_flag = 0;
	while ((ch = getopt(argc, argv, "hi")) != -1) {
		switch (ch) {
		case 'i':
			invert_flag = 1;
			break;
		case 'h':
		default:
			usage(progname);
			exit(ch == 'h' ? 0 : 1);
		}
	}
	argc -= optind;
	argv += optind;

	apm_fd = open("/dev/apm", O_RDONLY);
	if (apm_fd == -1)
		errx(1, "Cannot open APM device");
	memset(&current_info, 0, sizeof(current_info));

	error = init_tray(invert_flag);
	if (error == -1) {
		close(apm_fd);
		errx(1, "Cannot initialize notification area");
	}
	g_timeout_add(1000, (GSourceFunc)cb_timer, NULL);
	gtk_main();

	close(apm_fd);
	return (0);
}
