/*
 * Written by Alexey Vatchenko <av@bsdua.org>.
 * Public domain.
 */
#include <sys/param.h>
#include <sys/disklabel.h>
#include <sys/dkio.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/queue.h>
#include <sys/stat.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "tray-app.h"
#include "eject.xpm"

void	usage(const char *prog);

struct mntitem {
	char dev[MAXPATHLEN];
	char mp[MAXPATHLEN];

	SLIST_ENTRY(mntitem) next;
};

#define REMDEVICE_KEYLEN	64

struct remdevice {
	char key[REMDEVICE_KEYLEN];	/* device file name */
	SLIST_HEAD(, mntitem) mntlist;
	GtkWidget *menu_item;
	SLIST_ENTRY(remdevice) next;
};

struct ejector {
	SLIST_HEAD(, remdevice) devlist;
	GtkWidget *menu;
};

static int	alloc_mntitem(struct ejector *ejector, const char *key,
		    const char *label, struct remdevice **dev,
		    struct mntitem **item);
static void	free_devs(struct ejector *ejector);
static int	reset_ejector(struct ejector *ejector);
static void	inq2names(struct dk_inquiry *inq, char *name, size_t sz);
static void	disklabel2names(struct disklabel *dl, char *name);

void	ejector_init(struct ejector *ejector);
void	ejector_deinit(struct ejector *ejector);
int	ejector_enumerate(struct ejector *ejector);
void	ejector_popup_menu(struct ejector *ejector);
int	ejector_eject(struct remdevice *dev);

static void	prepare_tooltip(char *text, size_t sz);

static void	cb_menu_activate(gpointer data);
static gboolean	cb_tray_button_press(GtkStatusIcon *icon, GdkEventButton *event,
		    gpointer data);
static gboolean	cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
		    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data);

static int	init_menu(void);
static void	show_gui(void);

static int	init_tray(struct ejector *ejector, int doinvert);

static GtkStatusIcon	*tray_icon = NULL;
static GdkPixbuf	*tray_pixbuf = NULL;

void
usage(const char *prog)
{
	printf("Usage:\t%s [-h] [-i]\n"
	    "Options:\n"
	    "\t -h              - print this help\n"
	    "\t -i              - invert foreground color\n",
	    prog);
}

static int
alloc_mntitem(struct ejector *ejector, const char *key, const char *label,
    struct remdevice **dev, struct mntitem **item)
{
	struct remdevice *i, *rd, *last;

	rd = last = NULL;
	SLIST_FOREACH(i, &ejector->devlist, next) {
		if (strcmp(i->key, key) == 0) {
			rd = i;
			break;
		}
		last = i;
	}

	if (rd == NULL) {
		rd = (struct remdevice *)calloc(1, sizeof(*rd));
		if (rd == NULL)
			return (-1);
		strlcpy(rd->key, key, sizeof(rd->key));
		SLIST_INIT(&rd->mntlist);
		if (SLIST_EMPTY(&ejector->devlist))
			SLIST_INSERT_HEAD(&ejector->devlist, rd, next);
		else
			SLIST_INSERT_AFTER(last, rd, next);

		rd->menu_item = gtk_menu_item_new_with_label(label);
		if (rd->menu_item != NULL) {
			gtk_menu_shell_append(GTK_MENU_SHELL(ejector->menu),
			    rd->menu_item);
			gtk_widget_set_tooltip_text(GTK_WIDGET(rd->menu_item),
			    key);
			gtk_signal_connect_object(G_OBJECT(rd->menu_item),
			    "activate", G_CALLBACK(cb_menu_activate), rd);
		}
	}

	*item = (struct mntitem *)calloc(1, sizeof(**item));
	if (*item == NULL)
		return (-1);
	SLIST_INSERT_HEAD(&rd->mntlist, *item, next);
	*dev = rd;
	return (0);
}

static void
free_devs(struct ejector *ejector)
{
	struct remdevice *dev;
	struct mntitem *item;

	while (!SLIST_EMPTY(&ejector->devlist)) {
		dev = SLIST_FIRST(&ejector->devlist);
		SLIST_REMOVE_HEAD(&ejector->devlist, next);

		while (!SLIST_EMPTY(&dev->mntlist)) {
			item = SLIST_FIRST(&dev->mntlist);
			SLIST_REMOVE_HEAD(&dev->mntlist, next);
			free(item);
		}

		free(dev);
	}
}

static int
reset_ejector(struct ejector *ejector)
{
	if (ejector->menu != NULL)
		gtk_widget_destroy(ejector->menu);
	free_devs(ejector);
	ejector->menu = gtk_menu_new();
	return ((ejector->menu != NULL) ? 0 : -1);
}

/*
 * dk_inquiry's vendor is 64 bytes long, product is 128 bytes long (including
 * \0). Name is 192 bytes long. It's exactly what we need to store both names +
 * space between them.
 */
static void
inq2names(struct dk_inquiry *inq, char *name, size_t sz)
{
	strlcpy(name, inq->vendor, sz);
	strlcat(name, " ", sz);
	strlcat(name, inq->product, sz);
}

/*
 * disklabel's d_typename and d_un.un_d_packname each is 16 bytes long.
 * name is 192 bytes long. So, we don't need complex boundary checking.
 *
 */
static void
disklabel2names(struct disklabel *dl, char *name)
{
	size_t i, k;

	k = 0;
	for (i = 0; i < sizeof(dl->d_un.un_d_packname); i++)
		if (!isspace(dl->d_un.un_d_packname[i]))
			break;
	if (i < sizeof(dl->d_un.un_d_packname)) {
		k = sizeof(dl->d_un.un_d_packname) - i;
		memcpy(name, dl->d_un.un_d_packname + i, k);
		for (i = k; i > 0; i--) {
			if (!isspace(name[i - 1]) && name[i - 1] != '\0')
				break;
			k--;
		}
	}

	name[k] = '\0';
	if (*name != '\0')
		return;

	k = 0;
	for (i = 0; i < sizeof(dl->d_typename); i++)
		if (!isspace(dl->d_typename[i]))
			break;
	if (i < sizeof(dl->d_typename)) {
		k = sizeof(dl->d_typename) - i;
		memcpy(name, dl->d_typename + i, k);
		for (i = k; i > 0; i--) {
			if (!isspace(name[i - 1]) && name[i - 1] != '\0')
				break;
			k--;
		}
	}

	name[k] = '\0';
	if (*name == '\0')
		memcpy(name, "NONAME", 7);
}

void
ejector_init(struct ejector *ejector)
{
	memset(ejector, 0, sizeof(*ejector));
	SLIST_INIT(&ejector->devlist);
	ejector->menu = NULL;
}

void
ejector_deinit(struct ejector *ejector)
{
	if (ejector->menu != NULL)
		gtk_widget_destroy(ejector->menu);
	free_devs(ejector);
}

int
ejector_enumerate(struct ejector *ejector)
{
	char devname[MAXPATHLEN], key[REMDEVICE_KEYLEN];
	char vendprod[64 + 128];
	struct remdevice *dev;
	struct mntitem *item;
	struct statfs *mi;
	struct stat st;
	struct dk_inquiry inq;
	struct disklabel dl;
	int fd;
	int i, j, n, error;
	int isblk;
	size_t len;

	if (reset_ejector(ejector) == -1)
		return (-1);

	n = getmntinfo(&mi, 0);
	if (n == -1)
		return (-1);

	for (i = 0; i < n; i++) {
		error = stat(mi[i].f_mntfromname, &st);
		if (error == -1)
			continue;

		isblk = 0;
		strlcpy(devname, mi[i].f_mntfromname, sizeof(devname));
		len = strlen(devname);
		if (len == 0 || devname[len - 1] == '/')
			continue;
		if (S_ISBLK(st.st_mode)) {
			/* we need room for 'r' letter before device name */
			if (len >= MAXPATHLEN - 2)
				continue;

			for (j = len; j > 0; j--) {
				if (devname[j - 1] == '/') {
					memmove(devname + j + 1, devname + j,
					    len - j);
					devname[j] = 'r';
					devname[len + 1] = '\0';
					len++;
					isblk = 1;
					break;
				}
			}
			if (!isblk)
				continue;
		} else if (!S_ISCHR(st.st_mode))
			continue;

		fd = open(devname, O_RDONLY);
		if (fd == -1)
			continue;
		if (ioctl(fd, DIOCINQ, &inq) == 0) {
			/* removable devices (i.e. sd) */
			inq2names(&inq, vendprod, sizeof(vendprod));
		} else if (ioctl(fd, DIOCGDINFO, &dl) == 0) {
			/* the rest devices (e.g. cd) */
			disklabel2names(&dl, vendprod);
		} else {
			close(fd);
			continue;
		}
		close(fd);

		*key = '\0';
		for (j = len; j > 0; j--) {
			if (devname[j - 1] == '/') {
				strlcpy(key, devname + j + (isblk ? 1 : 0),
				    sizeof(key));
				break;
			}
		}
		len = strlen(key);
		if (len == 0)
			continue;
		key[len - 1] = '\0';

		alloc_mntitem(ejector, key, vendprod, &dev, &item);
		if (error == -1)
			continue;

		strlcpy(item->dev, mi[i].f_mntfromname, sizeof(item->dev));
		strlcpy(item->mp, mi[i].f_mntonname, sizeof(item->mp));
	}
	return (0);
}

void
ejector_popup_menu(struct ejector *ejector)
{
	gtk_widget_show_all(ejector->menu);
	gtk_menu_popup(GTK_MENU(ejector->menu), NULL, NULL,
	    gtk_status_icon_position_menu, tray_icon, 0,
	    gtk_get_current_event_time());
}

int
ejector_eject(struct remdevice *dev)
{
	GtkWidget *dialog;
	struct mntitem *item;
	char *reason;
	int fd;

	SLIST_FOREACH(item, &dev->mntlist, next) {
		if (unmount(item->mp, 0) == 0) {
			fd = open(item->dev, O_RDONLY);
			if (fd != -1) {
				ioctl(fd, DIOCEJECT);
				close(fd);
			}
		} else {
			reason = "";
			switch (errno) {
			case EPERM:
				reason = "\nYou don't have permissions.";
				break;
			case EBUSY:
				reason = "\nThe filesystem is busy.";
				break;
			}
			dialog = gtk_message_dialog_new(NULL, 0,
			    GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
			    "Cannot unmount \"%s\".%s", item->mp, reason);
			if (dialog != NULL) {
				gtk_dialog_run(GTK_DIALOG (dialog));
				gtk_widget_destroy(dialog);
			}
			return (-1);
		}
	}

	return (0);
}

static void
prepare_tooltip(char *text, size_t sz)
{
	strlcpy(text, "Safely Remove Hardware", sz);
}

static void
cb_menu_activate(gpointer data)
{
	ejector_eject((struct remdevice *)data);
}

static gboolean
cb_tray_button_press(GtkStatusIcon *icon, GdkEventButton *event, gpointer data)
{
	struct ejector *ej;

	ej = (struct ejector *)data;
	ejector_enumerate(ej);
	ejector_popup_menu(ej);
	return (TRUE);
}

static gboolean
cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data)
{
	char text[30];

	prepare_tooltip(text, sizeof(text));
	gtk_tooltip_set_text(tooltip, text);
	return (TRUE);
}

static int
init_tray(struct ejector *ejector, int doinvert)
{
	char tooltip[30];

	tray_pixbuf = gdk_pixbuf_new_from_xpm_data(eject_xpm);
	if (tray_pixbuf == NULL)
		return (-1);
	if (doinvert)
		trayapp_invert_pixbuf(tray_pixbuf);

	tray_icon = gtk_status_icon_new();
	if (tray_icon == NULL)
		return (-1);
	gtk_status_icon_set_from_pixbuf(tray_icon, tray_pixbuf);

	prepare_tooltip(tooltip, sizeof(tooltip));
	gtk_status_icon_set_tooltip_text(tray_icon, tooltip);

	g_signal_connect(tray_icon, "button-press-event",
	    G_CALLBACK(cb_tray_button_press), ejector);
	g_signal_connect(tray_icon, "query-tooltip",
	    G_CALLBACK(cb_tray_query_tooltip), ejector);
	return (0);
}

int
main(int argc, char **argv)
{
	int ch, error;
	int invert_flag;
	char *progname;
	struct ejector ej;

	progname = argv[0];

	if (!gtk_init_check(&argc, &argv))
		errx(1, "Cannot initialize windowing system");

	invert_flag = 0;

	while ((ch = getopt(argc, argv, "i")) != -1) {
		switch (ch) {
		case 'i':
			invert_flag = 1;
			break;
		default:
			usage(progname);
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	ejector_init(&ej);
	error = init_tray(&ej, invert_flag);
	if (error == -1)
		errx(1, "Cannot initialize notification area");
	gtk_main();
	return (0);
}
