/*
 * Written by Alexey Vatchenko <av@bsdua.org>.
 * Public domain.
 */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/audioio.h>

#include <gtk/gtk.h>

#include <err.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include "tray-app.h"
#include "sound.xpm"
#include "sound-0.xpm"
#include "sound-1.xpm"
#include "sound-2.xpm"

static void	usage(const char *prog);
static int	get_mixer_index(int fd, mixer_devinfo_t *devinfo);
static int	get_volume(int fd, mixer_devinfo_t *devinfo, u_char *volume);
static int	set_volume(int fd, mixer_devinfo_t *devinfo, u_char volume);
static int	get_mute(int fd, mixer_devinfo_t *devinfo, int *mute);
static int	set_mute(int fd, mixer_devinfo_t *devinfo, int mute);

static void	prepare_tooltip(int mute, u_char volume, char *text, size_t sz);

static gboolean	cb_timer(GtkWidget *widget);
static void	cb_button_toggled(GtkWidget *widget, gpointer data);
static void	cb_scale_value_changed(GtkScale *scale, GtkAdjustment *adj);
static gboolean	cb_window_delete(GtkWidget *widget,GdkEvent *event,
		    gpointer data);
static void	cb_window_destroy(GtkWidget *widget, gpointer data);
static gboolean	cb_window_key_press(GtkWidget *widget, GdkEventKey *event,
		    gpointer data);
static gboolean	cb_tray_button_press(GtkStatusIcon *icon, GdkEventButton *event,
		    gpointer data);
static gboolean	cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
		    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data);

static int	init_gui(void);
static void	show_gui(void);
static void	hide_gui(void);

static int	init_tray(int doinvert);
static void	set_tray_icon(u_char volume);

static int		mixer_fd;
static mixer_devinfo_t	outputs_master_dev[2];	/* volume, mute */

static GtkWidget	*gui_window = NULL;
static GtkObject	*gui_adj = NULL;
static GtkWidget	*gui_scale = NULL;
static GtkWidget	*gui_check = NULL;
static GtkWidget	*gui_box = NULL;

static GtkStatusIcon	*tray_icon = NULL;
static GdkPixbuf	*tray_pixbuf = NULL;
static GdkPixbuf	*tray_pixbuf_0 = NULL;
static GdkPixbuf	*tray_pixbuf_1 = NULL;
static GdkPixbuf	*tray_pixbuf_2 = NULL;

static u_char		current_volume = 0;
static int		current_mute = 0;

void
usage(const char *prog)
{
	printf("Usage:\t%s [-h] [-i]\n"
	    "Options:\n"
	    "\t -h              - print this help\n"
	    "\t -i              - invert foreground color\n",
	    prog);
}

static int
get_mixer_index(int fd, mixer_devinfo_t *devinfo)
{
	int error;
	int i, outputs_idx;

	i = 0;
	outputs_idx = -1;
	devinfo[0].index = 0;
	for (;;) {
		error = ioctl(fd, AUDIO_MIXER_DEVINFO, devinfo + i);
		if (error == -1)
			break;

		if (i == 0 && devinfo[0].type == AUDIO_MIXER_CLASS &&
		    strcmp(devinfo[0].label.name, "outputs") == 0)
			outputs_idx = devinfo[0].index;
		else if (i == 0 && devinfo[0].type == AUDIO_MIXER_VALUE &&
		    strcmp(devinfo[0].label.name, "master") == 0 &&
		    outputs_idx != -1 &&
		    devinfo[0].mixer_class == outputs_idx) {
			devinfo[1].index = devinfo[0].index;
			i++;
		} else if (i == 1 && devinfo[1].prev == devinfo[0].index &&
		    devinfo[1].type == AUDIO_MIXER_ENUM &&
		    strcmp(devinfo[1].label.name, "mute") == 0 &&
		    devinfo[1].mixer_class == outputs_idx)
			return (0);

		devinfo[i].index++;
	}
	return (-1);
}

static int
get_volume(int fd, mixer_devinfo_t *devinfo, u_char *volume)
{
	mixer_ctrl_t mctl;
	int error;

	memset(&mctl, 0, sizeof(mctl));
	mctl.dev = devinfo->index;
	mctl.type = AUDIO_MIXER_VALUE;
	error = ioctl(fd, AUDIO_MIXER_READ, &mctl);
	if (error == -1)
		return (-1);
	*volume = mctl.un.value.level[0];
	return (0);
}

static int
set_volume(int fd, mixer_devinfo_t *devinfo, u_char volume)
{
	mixer_ctrl_t mctl;
	int i, error;

	memset(&mctl, 0, sizeof(mctl));
	mctl.dev = devinfo->index;
	mctl.type = devinfo->type;
	mctl.un.value.num_channels = devinfo->un.v.num_channels;
	for (i = 0; i < devinfo->un.v.num_channels; i++)
		mctl.un.value.level[i] = volume;
	error = ioctl(fd, AUDIO_MIXER_WRITE, &mctl);
	if (error == -1)
		return (-1);
	return (0);
}

static int
get_mute(int fd, mixer_devinfo_t *devinfo, int *mute)
{
	mixer_ctrl_t mctl;
	int error;

	memset(&mctl, 0, sizeof(mctl));
	mctl.dev = devinfo->index;
	mctl.type = AUDIO_MIXER_ENUM;
	error = ioctl(fd, AUDIO_MIXER_READ, &mctl);
	if (error == -1)
		return (-1);
	*mute = mctl.un.ord;
	return (0);
}

static int
set_mute(int fd, mixer_devinfo_t *devinfo, int mute)
{
	mixer_ctrl_t mctl;
	int error;

	memset(&mctl, 0, sizeof(mctl));
	mctl.dev = devinfo->index;
	mctl.type = devinfo->type;
	mctl.un.ord = mute;
	error = ioctl(fd, AUDIO_MIXER_WRITE, &mctl);
	if (error == -1)
		return (-1);
	return (0);
}

static void
prepare_tooltip(int mute, u_char volume, char *text, size_t sz)
{
	if (mute) {
		strlcpy(text, "Audio is muted", sz);
	} else {
		snprintf(text, sz, "Audio volume: %u%%",
		    100U * (u_int)volume / AUDIO_MAX_GAIN);
	}
}

static gboolean
cb_timer(GtkWidget *widget)
{
	int mute;
	u_char volume;

	if (get_mute(mixer_fd, outputs_master_dev + 1, &mute) == -1)
		return (TRUE);
	else if (get_volume(mixer_fd, outputs_master_dev, &volume) == -1)
		return (TRUE);
	if (mute != current_mute || volume != current_volume) {
		set_tray_icon(mute ? 0 : volume);

		/* Move slider. Change "check" button state. */
		if (widget->window != NULL) {
			gtk_adjustment_set_value(GTK_ADJUSTMENT(gui_adj),
			    AUDIO_MAX_GAIN - volume);
			gtk_toggle_button_set_active(
			    GTK_TOGGLE_BUTTON(gui_check), mute);
			gtk_widget_set_sensitive(GTK_WIDGET(gui_scale), !mute);
		}

		current_volume = volume;
		current_mute = mute;
	}

	return (TRUE);
}

static void
cb_button_toggled(GtkWidget *widget, gpointer data)
{
	int mute;

	mute = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
	if (set_mute(mixer_fd, outputs_master_dev + 1, mute) == 0) {
		gtk_widget_set_sensitive(GTK_WIDGET(gui_scale), !mute);
		set_tray_icon(mute ? 0 : current_volume);
		current_mute = mute;
	}
}

static void
cb_scale_value_changed(GtkScale *scale, GtkAdjustment *adj)
{
	u_char volume;

	volume = AUDIO_MAX_GAIN - (u_char)gtk_adjustment_get_value(adj);
	if (set_volume(mixer_fd, outputs_master_dev, volume) == 0) {
		set_tray_icon(current_mute ? 0 : volume);
		current_volume = volume;
	}
}

static gboolean
cb_window_delete(GtkWidget *widget,GdkEvent *event, gpointer data)
{
	return (TRUE);	/* TRUE won't allow to destroy the window */
}

static void
cb_window_destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

static gboolean
cb_window_key_press(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if (event->hardware_keycode == 9 /* Escape */ ||
	    event->hardware_keycode == 36 /* Enter */)
		hide_gui();
	return (FALSE);
}

static gboolean
cb_tray_button_press(GtkStatusIcon *icon, GdkEventButton *event, gpointer data)
{
	if (gtk_widget_get_visible(gui_window))
		hide_gui();
	else
		show_gui();
	return (TRUE);
}

static gboolean
cb_tray_query_tooltip(GtkStatusIcon *icon, gint x, gint y,
    gboolean keyboard_mode, GtkTooltip *tooltip, gpointer data)
{
	char text[30];

	prepare_tooltip(current_mute, current_volume, text, sizeof(text));
	gtk_tooltip_set_text(tooltip, text);
	return (TRUE);
}

static int
init_gui(void)
{
	GtkRequisition req;
	int height, width;

	gui_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (gui_window == NULL)
		return (-1);
	g_signal_connect(gui_window, "delete-event",
	    G_CALLBACK(cb_window_delete), NULL);
	g_signal_connect(gui_window, "destroy", G_CALLBACK(cb_window_destroy),
	    NULL);
	g_signal_connect(gui_window, "key_press_event",
	    G_CALLBACK(cb_window_key_press), NULL);
	gtk_window_stick(GTK_WINDOW(gui_window));
	gtk_window_set_type_hint(GTK_WINDOW(gui_window),
	    GDK_WINDOW_TYPE_HINT_UTILITY);
	gtk_window_set_deletable(GTK_WINDOW(gui_window), FALSE);
	gtk_window_set_decorated(GTK_WINDOW(gui_window), FALSE);

	gui_adj = gtk_adjustment_new(0.0, AUDIO_MIN_GAIN,
	    AUDIO_MAX_GAIN, 1.0, 10.0, 0.0);
	if (gui_adj == NULL)
		return (-1);

	gui_check = gtk_check_button_new_with_label("mute");
	if (gui_check == NULL)
		return (-1);
	g_signal_connect(gui_check, "toggled", G_CALLBACK(cb_button_toggled),
	    NULL);

	gui_scale = gtk_vscale_new(GTK_ADJUSTMENT(gui_adj));
	if (gui_window == NULL)
		return (-1);
	gtk_widget_set_size_request(GTK_WIDGET(gui_scale), -1, 200);
	gtk_scale_set_draw_value(GTK_SCALE(gui_scale), FALSE);
	g_signal_connect(gui_scale, "value_changed",
	    G_CALLBACK(cb_scale_value_changed), gui_adj);

	/*
	 * Set size of the window according to the summary sizes of all
	 * window's objects
	 */
	height = width = 0;
	gtk_widget_size_request(gui_scale, &req);
	height += req.height;
	width += req.width;
	gtk_widget_size_request(gui_check, &req);
	height += req.height;
	width += req.width;
	height += 5; /* spacing */
	gtk_widget_set_size_request(gui_window, width, height);

	gui_box = gtk_vbox_new(FALSE, 5);
	if (gui_box == NULL)
		return (-1);
	gtk_box_pack_start(GTK_BOX(gui_box), GTK_WIDGET(gui_scale), FALSE,
	    FALSE, 0);
	gtk_box_pack_start(GTK_BOX(gui_box), GTK_WIDGET(gui_check), FALSE,
	    FALSE, 0);

	gtk_container_add(GTK_CONTAINER(gui_window), gui_box);
	return (0);
}

void
show_gui(void)
{
	GdkRectangle area;
	GtkOrientation orientation;
	int width, height, x, y;

	gtk_adjustment_set_value(GTK_ADJUSTMENT(gui_adj),
	    AUDIO_MAX_GAIN - current_volume);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gui_check),
	    current_mute);
	gtk_widget_set_sensitive(GTK_WIDGET(gui_scale), !current_mute);
	gtk_widget_show_all(GTK_WIDGET(gui_window));

	gtk_status_icon_get_geometry(tray_icon, NULL, &area, &orientation);
	gtk_window_get_size(GTK_WINDOW(gui_window), &width, &height);
	x = area.x;
	y = area.y;
	if (orientation == GTK_ORIENTATION_HORIZONTAL) {
		if (x + width > gdk_screen_width())
			x = x + area.width - width;
		if (y + area.height + height > gdk_screen_height())
			y -= height;
		else
			y += area.height;
	} else {
		if (x + area.width + width > gdk_screen_width())
			x -= width;
		else
			x += area.width;
		if (y + height > gdk_screen_height())
			y = y + area.height - height;
	}
	gtk_window_move(GTK_WINDOW(gui_window), x, y);
}

void
hide_gui(void)
{
	gtk_widget_hide(GTK_WIDGET(gui_window));
}

static int
init_tray(int doinvert)
{
	char tooltip[30];
	u_char volume;
	int mute;

	tray_pixbuf = gdk_pixbuf_new_from_xpm_data(sound_xpm);
	if (tray_pixbuf == NULL)
		return (-1);
	tray_pixbuf_0 = gdk_pixbuf_new_from_xpm_data(sound_0_xpm);
	if (tray_pixbuf_0 == NULL)
		return (-1);
	tray_pixbuf_1 = gdk_pixbuf_new_from_xpm_data(sound_1_xpm);
	if (tray_pixbuf_1 == NULL)
		return (-1);
	tray_pixbuf_2 = gdk_pixbuf_new_from_xpm_data(sound_2_xpm);
	if (tray_pixbuf_2 == NULL)
		return (-1);

	if (doinvert) {
		trayapp_invert_pixbuf(tray_pixbuf);
		trayapp_invert_pixbuf(tray_pixbuf_0);
		trayapp_invert_pixbuf(tray_pixbuf_1);
		trayapp_invert_pixbuf(tray_pixbuf_2);
	}

	tray_icon = gtk_status_icon_new();
	if (tray_icon == NULL)
		return (-1);
	volume = 0;
	mute = 0;
	get_volume(mixer_fd, outputs_master_dev, &volume);
	get_mute(mixer_fd, outputs_master_dev + 1, &mute);
	set_tray_icon(mute ? 0 : volume);
	current_volume = volume;
	current_mute = mute;

	prepare_tooltip(mute, volume, tooltip, sizeof(tooltip));
	gtk_status_icon_set_tooltip_text(tray_icon, tooltip);

	g_signal_connect(tray_icon, "button-press-event",
	    G_CALLBACK(cb_tray_button_press), NULL);
	g_signal_connect(tray_icon, "query-tooltip",
	    G_CALLBACK(cb_tray_query_tooltip), NULL);
	return (0);
}

static void
set_tray_icon(u_char volume)
{
	GdkPixbuf *pb;

	if (volume == 0)
		pb = tray_pixbuf_0;
	else if (volume < 100)
		pb = tray_pixbuf_1;
	else if (volume < 168)
		pb = tray_pixbuf_2;
	else
		pb = tray_pixbuf;
	gtk_status_icon_set_from_pixbuf(tray_icon, pb);
}

int
main(int argc, char **argv)
{
	char *progname;
	int ch, invert_flag;
	int error;

	progname = argv[0];

	if (!gtk_init_check(&argc, &argv))
		errx(1, "Cannot initialize windowing system");

	invert_flag = 0;
	while ((ch = getopt(argc, argv, "hi")) != -1) {
		switch (ch) {
		case 'i':
			invert_flag = 1;
			break;
		case 'h':
		default:
			usage(progname);
			exit(ch == 'h' ? 0 : 1);
		}
	}
	argc -= optind;
	argv += optind;

	mixer_fd = open("/dev/mixer", O_RDWR);
	if (mixer_fd == -1)
		errx(1, "Cannot open mixer device");

	error = get_mixer_index(mixer_fd, outputs_master_dev);
	if (error == -1) {
		close(mixer_fd);
		errx(1, "Cannot get mixer information");
	}

	error = init_tray(invert_flag);
	if (error == -1) {
		close(mixer_fd);
		errx(1, "Cannot initialize notification area");
	}
	error = init_gui();
	if (error == -1) {
		close(mixer_fd);
		errx(1, "Cannot initialize program window");
	}
	g_timeout_add(1000, (GSourceFunc)cb_timer, (gpointer)gui_window);
	gtk_main();

	close(mixer_fd);
	return (0);
}
